/* -------------------------------------------------------------------------- */
/*                                                                            */
/* (C) Copyright D.C.Devenport 1998. All right reserved.                      */
/*                                                                            */
/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY      */
/* KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE        */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR      */
/* PURPOSE.                                                                   */
/*                                                                            */
/* This code, and no part of this code, may not be used in any                */
/* commercial or for-profit venture without the express written               */
/* permission of D.C.Devenport. (BeebInC@aol.com)                             */
/*                                                                            */
/* Credit must be given within any program that uses any of this code         */
/* OR in the accompanying documentation. (And mail me a copy :) )             */
/*                                                                            */
/*----------------------------------------------------------------------------*/
#include <i86.h>
#include <conio.h>
#include <stdlib.h>
#include <stdio.h>
#include "6845crtc.h"
#include "bbc.h"
#include "sysvia.h"
#include "main.h"
#include "speech.h"
#include "matrix.h"
#include "sound.h"
#include "modex.h"
#include "mode7c.h"
#include "modexc.h"

int SystemInterruptPendingFlag;
WORD ScreenStart;

// 0xffff * 2 = 0x1fffe
#define TIMERWRAPVALUE 0x20000

// 8us at 2MHz is 16 cycles + a safety margin
#define SOUND_HOLD 16


// only 16 regs, but 4 are different on reads
static
BYTE ORB,IRB,
     ORA,IRA,
     DDRB, // 0 bit means input, a 1 output
     DDRA, // 0 bit means input, a 1 output
     T1CL, // read = t1 low order counter
     T1CH,
     T1LL,
     T1LH,
     T2CL, // read = t2 low order counter
     T2CH,
     T2LL,
     SR,
     ACR,
     PCR,
     IFR,
     IER,
     VSYNC_HOLD=0,
     OldLatch,OldSoundLatch,
     PortA,PortB; // these variables contain the actual values on the lines...

int VerticalSYNCTimer;
// these values are multiplied by 2, so they can be decremented at 2MHz
static int Timer1,Timer2,SoundTimer;
static BYTE SetInterruptOnTimer1,SetInterruptOnTimer2;


static BYTE LatchNOTEnabled[8];
static WORD ScreenStartArray[]= { 0x4000,0x6000,0x3000,0x5800 };
static BYTE ScreenStartIndex= 0;

#include "viddraw.c"


#define DOSOUND \
    if (OldSoundLatch==0) \
    { \
      UpdateClocksAndInterrupts(); \
        if (SoundTimer>=SOUND_HOLD) \
        { \
          SoundTimer=0;  \
          WriteSound(PortA); \
        } \
    } 


#define DOKEYBOARD \
    if (!LatchNOTEnabled[LATCH_KEYBOARDWRITE]) \
      PortA=(ReadKeyboard(PortA));

/*
#define DOKEYBOARD \
    if (!LatchNOTEnabled[LATCH_KEYBOARDWRITE]) \
    { \
      BYTE OldPortA=PortA; \
      IRA=PortA & (DDRA ^ 255); \
      PortA&=DDRA; \
      PortA|=(ReadKeyboard(OldPortA)); \
      IRA|=PortA & (DDRA ^ 255); \
    }
*/



void ResetSystemVIA()
{
  int Count;
    for (Count=0; Count<8; Count++)
      LatchNOTEnabled[Count]=8;

  DefaultVSYNCValue=DEFAULT_VSYNC;
  VerticalSYNCTimer=DEFAULT_VSYNC; // let ROM init properly
  VSYNCPhase=VSYNC_VSyncStart;
  VSYNCCount=0;
  VSYNC_HOLD=0;

  Timer1=TIMERWRAPVALUE;
  Timer2=TIMERWRAPVALUE;
  T1LL=0xff;
  T1LH=0xff;
  T2LL=0xff;

  PortB=0xf0; // fire button 0 and 1 off and no speech
  IRB=0xf0;

  IFR=0;
//  IER=0; // dont clear, otherwise always does hard reset
  SystemInterruptPendingFlag=0;
  InterruptPendingFlag=0;

  InterlaceTime=0;
}



BYTE ReadSystemVIA(WORD Address)
{
  BYTE Result;

// ASSUMPTION - all cycles upto the actual read have been done
  SyncClocks();  // its a slow device so sync with it

    switch (Address & 0xf) // now the correct register
    {
      case 0  : UpdateClocksAndInterrupts();
                Result=ORB & DDRB;
                  if ((ACR & 0x02)==0) // latching disabled
                    Result|=((DDRB ^ 255) & PortB);
                  else // latching enabled
                    Result|=((DDRB ^ 255) & IRB);

                  if ((PCR & 0xe0)==0x20 || (PCR & 0xe0)==0x60) // independant mode
                    IFR&=0xef; // clear bit 4
                  else
                    IFR&=0xe7; // clear bits 3&4

                Result|=0xc0; // no speech frig up
                
                return (Result);
                break;
      case 1  : // IRQ handshaking
                UpdateClocksAndInterrupts();
                   if ((PCR & 0x0e)==2 || (PCR & 0x0e)==6) // independant mode
                     IFR&=0xfd; // clear bit 1
                   else           
                     IFR&=0xfc; // clear bits 0 & 1
                
                // fall thru
      case 15 : Result=ORA & DDRA;

//                  if (!LatchNOTEnabled[LATCH_SPEECHREAD])
//                    PortA=ReadSpeech();

//                  if ((ACR & 0x01)==0) // latching disabled
//                    Result|=((DDRA ^ 255) & PortA);
//                  else // latching enabled
//                    Result|=((DDRA ^ 255) & IRA);

                    Result|=((DDRA ^ 255) & PortA);
                return (Result);
                break;
      case 2  : return (DDRB);
                break;
      case 3  : return (DDRA);
                break;
      case 4  : UpdateClocksAndInterrupts();
                IFR&=0xbf; // Bit 6 cleared in IFR
                SetInterruptOnTimer1=TRUE;
                T1CL=(BYTE) (Timer1>>1);
                return (T1CL);
                break;
      case 5  : UpdateClocksAndInterrupts();
                T1CH=(BYTE) (Timer1>>9);
                return (T1CH);
                break;
      case 6  : return (T1LL);
                break;
      case 7  : return (T1LH);
                break;
      case 8  : UpdateClocksAndInterrupts();
                IFR&=0xdf; // bit 5 cleared in IFR
                T2CL=(BYTE) (Timer2>>1);
                return (T2CL);
                break;
      case 9  : UpdateClocksAndInterrupts();
                T2CH=(BYTE) (Timer2>>9);
                return (T2CH);
                break;
      case 10 : UpdateClocksAndInterrupts();
                IFR&=0xfb; // clear bit 2
                return (SR); // SR does nothing
                break;
      case 11 : return (ACR);
                break;
      case 12 : return (PCR);
                break;
      case 13 : // unset/set bit 7 before read
                  if ((IFR & IER & 0x7f)==0)
                    IFR&=0x7f;
                  else
                    IFR|=0x80;
                return (IFR | VSYNC_HOLD);
                break;

      case 14 : return (128 | IER); // bit 7 always set on read
                break;
    }
  return 0;
}


void WriteSystemVIA(WORD Address,BYTE Value)
{
// ASSUMPTION - all cycles upto the actual write have been done
  SyncClocks();  // its a slow device so sync with it

    switch (Address & 0xf) // now the correct register
    {
      case 0   : UpdateClocksAndInterrupts();
                 ORB=Value;
                  if ((PCR & 0xe0)==0x20 || (PCR & 0xe0)==0x60) // independant mode
                    IFR&=0xef; // clear bit 4
                  else
                    IFR&=0xe7; // clear bits 3 & 4
                
      NewPortB :
                PortB=(PortB & (DDRB ^ 255)) + (ORB & DDRB);

                OldLatch=LatchNOTEnabled[PortB & 0x7];
                LatchNOTEnabled[PortB & 0x7]=(PortB & 0x8);
                  switch (PortB & 0x0f)
                  {
                    case 0x4 : ScreenStartIndex&=2;
                               ScreenStart=ScreenStartArray[ScreenStartIndex];
                               break;
                    case 0xc : ScreenStartIndex|=1;
                               ScreenStart=ScreenStartArray[ScreenStartIndex];
                               break;

                    case 0x5 : ScreenStartIndex&=1;
                               ScreenStart=ScreenStartArray[ScreenStartIndex];
                               break;
                    case 0xd : ScreenStartIndex|=2;
                               ScreenStart=ScreenStartArray[ScreenStartIndex];
                               break;

                    case 0x6 : CAPSLockOn();
                               break;
                    case 0xe : CAPSLockOff();
                               break;
                    case 0x7 : SHIFTLockOn();
                               break;
                    case 0xf : SHIFTLockOff();
                               break;

                    case 0x0 : // sound write enable enabled
                                 if (OldSoundLatch)
                                 {
                                   UpdateClocksAndInterrupts();
                                   SoundTimer=0;
                                   OldSoundLatch=0;
                                 }
                                 else
                                 {
                                   DOSOUND;
                                 }
                               break;
                    case 0x8 : // sound write enable disabled
                               DOSOUND;
                               OldSoundLatch=8;
                               break;
                    case 0x1 : // read select on speech, on
                               break;
                    case 0x9 : // read select on speech, off
                               break;
                    case 0x2 : // write select on speech, on
                               break;
                    case 0xa : // write select on speech, off
                               break;
                    case 0x3 : // keyboard write enable, on
                               break;
                    case 0xb : // keyboard write enable, off
                               break;
                    default : ;// nothing
                  }

                ModeXScreenWrapRestart=(DWORD) (&AddressSpace[ScreenStart]);
                break;

      case 1  : // IRQ handshaking
                UpdateClocksAndInterrupts();
                   if ((PCR & 0x0e)==2 || (PCR & 0x0e)==6) // independant mode
                     IFR&=0xfd; // clear bit 1
                   else
                     IFR&=0xfc; // clear bits 0 & 1
                // fall thru
      case 15 : // same as case 1 - no handshaking interrupt
                ORA=Value;
      NewPortA :
                PortA=(PortA & (DDRA ^ 255)) + (ORA & DDRA);
//                  if (!LatchNOTEnabled[LATCH_SPEECHWRITE])
//                    WriteSpeech(PortA);

                DOKEYBOARD;
//                DOSOUND;
                break;
      case 2  : DDRB=Value;
                break;
                goto NewPortB; // change line status for port
                break;
      case 3  : DDRA=Value;
                break;
                goto NewPortA; // change line status for port
      case 4  : T1LL=Value;
                break;
      case 5  : UpdateClocksAndInterrupts();

                T1LH=Value;
                T1CL=T1LL;
                T1CH=T1LH;
                IFR&=0xbf; // Bit 6 cleared in IFR

                  if ((ACR & 0xc0)==0x80 && (DDRB & 0x80)==0x80) // one shot mode
                    PortB&=0x7f; // PB7 goes lo (presumably logic 0)

                Timer1=(T1CH*256 + T1CL)*2+INITIALTIMERADDVALUE;
                SetInterruptOnTimer1=TRUE;

                break;
      case 6  : T1LL=Value;
                break;
      case 7  : //  if ((ACR & 0xc0)==0x80) // one shot mode
                //    return; // no effect in one shot mode
                T1LH=Value;
                break;
      case 8  : T2LL=Value;
                break;
      case 9  : UpdateClocksAndInterrupts();
                T2CH=Value;
                T2CL=T2LL;
                IFR&=0xdf; // clear bit 5

                Timer2=(T2CH*256 + T2CL)*2+INITIALTIMERADDVALUE;
                SetInterruptOnTimer2=TRUE;
                break;
                
      case 10 : UpdateClocksAndInterrupts();
                IFR&=0xfb; // clear bit 2
                SR=Value;
                break;
      case 11 : ACR=Value;
                break;
      case 12 :   if ( ((PCR & 0xe0)==0xc0) && ((Value & 0xe0)==0xe0) )
                  { // pulsed LPSTB high - so fake a lightpen interrupt
                    int PenTimer;

                    UpdateClocksAndInterrupts();

                    PenTimer=GetScanLineValue();

//                    SetSysCB2(); // a fix?
                    // 128 cycles per screen row...
                    RegisterValues6845[0x10]=((BYTE) (PenTimer>>7) & 0x3f);
                    RegisterValues6845[0x11]=((BYTE) (PenTimer<<1) );
                  }
                PCR=Value;
                break;
      case 13 : UpdateClocksAndInterrupts();
                Value&=0x7f; // bit 7 cannot be just 'cleared'
                IFR&=(Value ^ 0x7f) ;  // clear 'set' bits
                  if ((IFR & IER & 0x7f)>0)
                    IFR|=0x80;
                break;
      case 14 : UpdateClocksAndInterrupts();
                  if (Value>=128) // setting bits
                    IER|=(Value & 0x7f);
                  else // clearing bits
                    IER&=(Value ^ 0x7f);
                break;
    }
}


int SysViaMinClock()
{
#ifdef DEV
  printf("VSYNC : %d, T1=%d, T2=%d\n",VerticalSYNCTimer,Timer1,Timer2);
#endif

    if (Timer1<Timer2)
    {
        if (VerticalSYNCTimer<Timer1)
          return (VerticalSYNCTimer);
        else
          return (Timer1);
    }
    else // timer2<Timer1
    {
        if (VerticalSYNCTimer<Timer2)
          return (VerticalSYNCTimer);
        else
          return (Timer2);
    }
}


void CalcSystemVia(int Cycles)
{
  // GENERATE Interrupts if necessary  
// PB7 goes hi on T1 time out (single shot mode)
// IFR bit set

  SoundTimer+=Cycles;

  //decrement Timer1
  Timer1-=Cycles;
    if (Timer1<TIMEDOUT)
    {
        switch (ACR & 0xc0)
        {
           case 0x80 :  // one shot mode PB7 goes hi (presumably logic 1)
                          if (DDRB & 0x80==0x80)
                            PortB|=0x80; // 
                        // fall thru
           case 0x00 :  Timer1+=TIMERWRAPVALUE+TIMERADDVALUE;

                          if (SetInterruptOnTimer1)
                          {
                            IFR|=64; // t1 interrupt flag set
                            SetInterruptOnTimer1=FALSE;
                          }
                        break;

           case 0xc0 :  // continuous mode
                          if (DDRB & 0x80==0x80)
                            PortB^=0x80;
                        // fall thru
           case 0x40 :  Timer1+=(((T1LH<<8) + T1LL)<<1)+TIMERADDVALUE;
                        IFR|=64; // t1 interrupt flag set
                        break;
        }
      UpdateSound(); // envelopes processed on timer1
    }

  // test timer2 action
    if (ACR & 0x20==0) // one shot mode
    {
     // T2 time out (1 shot mode) disables its interrupt
     //decrement Timer2
      Timer2-=Cycles;
        if (Timer2<TIMEDOUT)
        {
          Timer2+=TIMERWRAPVALUE+TIMERADDVALUE;

            if (SetInterruptOnTimer2)
            {
              IFR|=32; // t2 interrupt flag set
              SetInterruptOnTimer2=FALSE;
            }
        }
    }
/*
    else // pulse counting mode
    {
      // do nothing unless PB6 is pulsed - speech assume no pulse
      // then decrement T2
        if (Timer2<0)
        {
          Timer2+=TIMERWRAPVALUE+TIMERADDVALUE;
          IFR|=32; // t2 interrupt flag set
        }

    }
*/

#ifdef __VSYNCERROR__
  VSYNCError+=Cycles;
#endif

  //  Decrement VerticalSYNCTimer
  VerticalSYNCTimer-=Cycles;
    if (VerticalSYNCTimer<=0)
      DoVideoStuff();

    if ((IFR & IER & 0x7f)>0)  // set interrupt flag in IFR
    {
      IFR|=0x80;
      SystemInterruptPendingFlag=TRUE;
    }
    else // clear interrupt flag in IFR
    {
      SystemInterruptPendingFlag=FALSE;
      IFR&=0x7f;
    }
}


void SetSysCA2(BYTE PositiveFlag)
{
    if (PositiveFlag>1) // keyboard read test
    {
      KeyboardUpdateClocksAndInterrupts();
      IFR|=0x1; // set CA2 interrupt
      return;
    }

    switch (PCR & 0x0e)
    {
      case 0 :
      case 2 :
                 if (!PositiveFlag)
                 {
                   KeyboardUpdateClocksAndInterrupts();
                   IFR|=0x1; // set CA2 interrupt
                 }
               break;
      case 4 :
      case 6 :
                 if (PositiveFlag)
                 {
                   KeyboardUpdateClocksAndInterrupts();
                   IFR|=0x1; // set CA2 interrupt
                 }
               break;
    }
}


void SetSysCB1(BYTE Button1,BYTE Button2)  // EOC from ADC
{ 
  UpdateClocksAndInterrupts();
  IFR|=0x10; // set CB1 interrupt
    if (Button1)
      PortB&=0xef;
    else
      PortB|=0x10;

    if (Button2)
      PortB&=0xdf;
    else
      PortB|=0x20;

  IRB=(PortB & (DDRB ^ 255)) + (ORB & DDRB);
}


// lightpen - never occurs extenally
void SetSysCB2()
{
  UpdateClocksAndInterrupts();
  IFR|=0x8; // set CB2 interrupt
}



void DumpSysVIA(FILE * FileHandle)
{

  UpdateClocksAndInterrupts();
  fprintf(FileHandle,"Slow Bus %d %d %d %d %d %d %d %d\n",
                      LatchNOTEnabled[0],LatchNOTEnabled[1],
                      LatchNOTEnabled[2],LatchNOTEnabled[3],
                      LatchNOTEnabled[4],LatchNOTEnabled[5],
                      LatchNOTEnabled[6],LatchNOTEnabled[7]);
  fprintf(FileHandle,"ORB = %2x  IRB = %2x  DDRB = %2x\n",ORB,IRB,DDRB);
  fprintf(FileHandle,"ORA = %2x  IRA = %2x  DDRA = %2x\n",ORA,IRA,DDRA);
  fprintf(FileHandle,"T1CL = %2x  T1CH = %2x\n",T1CL,T1CH);
  fprintf(FileHandle,"T1LL = %2x  T1LH = %2x\n",T1LL,T1LH);
  fprintf(FileHandle,"T2CL = %2x  T2CH = %2x\n",T2CL,T2CH);
  fprintf(FileHandle,"T2LL = %2x\n",T2LL);
  fprintf(FileHandle,"SR   = %2x\n",SR);
  fprintf(FileHandle,"ACR  = %2x\n",ACR);
  fprintf(FileHandle,"PCR  = %2x\n",PCR);
  fprintf(FileHandle,"IFR  = %2x  IER  = %2x\n",IFR,IER);
  fprintf(FileHandle,"Port A = %2x  Port B = %2x\n",PortA,PortB);

  fprintf(FileHandle,"Timer1 = %d  Timer2 = %d\n",Timer1,Timer2);
  fprintf(FileHandle,"VSYNC Timer = %d\n",VerticalSYNCTimer);
  fprintf(FileHandle,"DefaultVSYNCValue = %d   VSYNCPhase = %d   VSYNCCount = %d\n",
                      DefaultVSYNCValue,VSYNCPhase,VSYNCCount);
#ifdef __VSYNCERROR__
  fprintf(FileHandle,"VSYNC Error = %d\n\n",
          VSYNCError);
#endif
}
